#include "cmdLineParser.h"

int Parser::isOption(const char *arg){
	if(arg == NULL || strlen(arg) < 2 )
		return 0;

   	if(arg[0] == '-' && !isdigit(arg[1]))
      	return 1;
   	return 0;
}

void Parser::replaceString(string& subject, const string& search, const string& replace){
    size_t pos = 0;
    while ((pos = subject.find(search, pos)) != std::string::npos) {
         subject.replace(pos, search.length(), replace);
         pos += replace.length();
    }
}

void Parser::cmdLineHelp(){
	cerr << "Compass Options: \n" ;
	for( int i = 0 ; i < helpOpt.size() ; i++){
		cerr.setf(ios::left);
        replaceString(helpOpt[i][2], "\n", "\n       ");
		cerr << "    " <<  helpOpt[i][0];
	    cerr << " " << helpOpt[i][1]  << "\n       ";
		cerr << helpOpt[i][2] << endl;
	}
}

int Parser::inputParser(char** optn, const string& argv){
    string errormsg = "Compass: Unable to associate the argument \'" + argv + "\' with any valid options!\n Try \'compass --help\' for more information.\n";

    if(!strcmp(*optn, "-n")){
        nullModel = argv;
        return 1;
    }
    if(!strcmp(*optn, "-d")){
        compModel = argv;
        return 1;
    }

    if(!strcmp(*optn, "-p")){
        pflag = true;
        sourcePath = argv;
        return 1;
    } 
    if(!strcmp(*optn, "-f")){
        fflag = true;
        sourceFile = argv;
        return 1;
    } 

    if(!strcmp(*optn, "-c")){
        if(argv == "ON" || argv == "on")
            checkpoint = true; 
        else checkpoint = false;
        return 1;    
    }
    if(!strcmp(*optn, "-u")){
        if(argv == "ON" || argv == "on")
            usageFile = true; 
        else checkpoint = false;    
        return 1;       
    }
    if(!strcmp(*optn, "-s")){
        inptDictionary = argv; 
        return 1; 
    }    
    
    if(!strcmp(*optn, "-o")){
        if(argv == "none")
            outFlag = 0;
        else if(argv == "minimal")
            outFlag = 1;
        else if(argv == "basic")
            outFlag = 2;
        else if(argv == "full")
            outFlag = 3;
        else {
            cerr << errormsg;
            exit(0);
        }  
        return 1;     
    }

    if(!strcmp(*optn, "-l")){
        if(argv == "none")
            logFlag = 0;
        else if(argv == "minimal")
            logFlag = 1;
        else if(argv == "basic")
            logFlag = 2;
        else if(argv == "full")
            logFlag = 3;
        else {
            cerr << errormsg;
            exit(0);
        }   
        return 1;       
    }

    if(!strcmp(*optn, "-T")){
        try {
            kelvin = stod(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1;   
    }
    if(!strcmp(*optn, "-S")){
        try {
            stoppingKelv = stod(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1;   
    }
    if(!strcmp(*optn, "-P")){
        try {
            kelvProb = stod(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1;   
    }
    if(!strcmp(*optn, "-I")){
        try {
            perturbations = stoi(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1;   
    }
    if(!strcmp(*optn, "-E")){
        try {
            tolerance = stod(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1;   
    }
    if(!strcmp(*optn, "-delta")){
        try {
            delta = stoi(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1; 
    }
    if(!strcmp(*optn, "-epsln")){
        try {
            epsln = stoi(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1; 
    }
    if(!strcmp(*optn, "-UB")){
        try {
            upperBound = stoi(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1; 
    }
    if(!strcmp(*optn, "-LB")){
        try {
            lowerBound = stoi(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1; 
    }
    if(!strcmp(*optn, "-ID")){
        try {
            updateLen = stoi(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1; 
    }
    if(!strcmp(*optn, "-R")){
        try {
            regionLen = stod(argv) / log2(exp(1));
        }
        catch(const std::exception& e) {
           cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1; 
    }

    if(!strcmp(*optn, "-SUB")){
        try {
            seqUpper = stoi(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1; 
    }
    if(!strcmp(*optn, "-SLB")){
        try {
            seqLower = stoi(argv);
        }
        catch(const std::exception& e) {
            cerr << errormsg;
            cerr << e.what() << '\n';
            exit(0);
        }
        return 1; 
    }

    if(!strcmp(*optn, "-disectLegnd")){
        if(argv == "ON" || argv == "on")
            disectLegend = true; 
        return 1;     
    }
    if(!strcmp(*optn, "-pertLegnd")){
        if(argv == "ON" || argv == "on")
            pertLegend = true; 
        return 1;     
    }
    if(!strcmp(*optn, "-seqDisect")){
        if(argv == "ON" || argv == "on")
            seqDisect = true; 
        return 1;      
    }
    if(!strcmp(*optn, "-seqCost")){
        if(argv == "ON" || argv == "on")
            seqCost = true; 
        return 1;     
    }

    return 0;
} 


Parser::Parser(){}

Parser::Parser(int argc, char **argv){
    bool isOptn = true;
    char ** optn;
    
    if(argc == 2 && !strcmp(argv[1],"--help")){
        cmdLineHelp();
        exit(0); 
    } 

    for(int i = 1; i < argc; i++){
        if(isOption(argv[i])){
            optn = &argv[i];
            isOptn = false;
        } else if(isOptn){
            cerr << "Compass: Unable to associate the argument \'" << argv[i] << "\' with any valid options!\n";
            cerr << "Try \'compass --help\' for more information.\n";
            exit(0);
        } else {
            if(inputParser(optn, argv[i]))
                isOptn = true;
            else {
                cerr << "Compass: Unable to associate the argument \'" << argv[i] << "\' with any valid options!\n";
                cerr << "Try \'compass --help\' for more information.\n";
                exit(0);
            }
        }
    } 

    if(!((pflag || fflag) && (!pflag || !fflag))){
        cerr << "Compass: Cannot use both or none of the -p and -f options in conjunction.\n";
        cerr << "Use either one of the -p or -f options.\n";
        exit(0);
    }

    if(sourceFile == "" && sourcePath == ""){
        cerr << "Compass: Unable to associate the input source collection.\n";
        cerr << "Use either one of the -p or -f options to input the source collection.\n";
        cerr << "Try \'compass --help\' for more information.\n";
        exit(0) ;
    }

    if(!(checkpoint >= 0 && checkpoint < perturbations)){
        cerr << "Compass: Invalid count for checkpoint.\n";
        cerr << "Input an integer between [0,number of perturbations]\n";
        cerr << "Try \'compass --help\' for more information.\n"; 
        exit(0) ;
    }

    if(outFlag == 3) {
        pertLegend = true;
        disectLegend = true;
    }    

    if(inptDictionary != ""){
        cpData = file_read(inptDictionary);
        if(cpData.size() < 2){
            cerr << "Compass: CheckPoint Error: Not enough data to check point.'" << endl;
            exit(0);
        }
        istringstream iss(cpData[0]);
        vector<string> splitRes((istream_iterator<string>(iss)),
                                 istream_iterator<string>());
        if(splitRes.size() != 11){
            cerr << "Compass: CheckPoint Error: Not enough input parameters to check point.'" << endl;
            exit(1);
        }                         
        kelvin = stod(splitRes[0]);
        kelvProb = stod(splitRes[1]);
        stoppingKelv = stod(splitRes[2]);
        perturbations = stoi(splitRes[3]);
        
        lowerBound = stoi(splitRes[4]);
        upperBound = stoi(splitRes[5]);
        updateLen = stoi(splitRes[6]);
        iter = stoi(splitRes[7]);
        delta = stoi(splitRes[8]);
        epsln = stoi(splitRes[9]);
        checkpointedML = stod(splitRes[10]);

        assert(lowerBound >= 3 && lowerBound <= 100);
        assert(upperBound > lowerBound && upperBound <= 100);
        assert(updateLen > 1 && updateLen <= 10);
        assert(delta >= 5 && delta <= 50000);
        assert(epsln >= 5 && epsln <= 50000);
        assert(delta <= epsln);
        assert(checkpointedML > 0);
        assert(kelvin > 0 && kelvProb < 1 && kelvProb > 0 && kelvin > stoppingKelv && perturbations > 0);
    }   
}


