#ifndef COST_H
#define COST_H
#include <vector>
#include <string>
#include <algorithm>
using namespace std;

class Cost{
public:    
    vector<double> matrix;
    int nSeq = 0;
    int delta = 0;

    Cost(const int& n, const int& nCcpt, const int& delt){
        assert(n > 0);
        assert(nCcpt >= 0);
        nSeq = n;
        delta = delt;
        int vecSize = 0;

        if(nCcpt-delta > nSeq) vecSize = 1;
        else if(nCcpt == 0 || (nCcpt-delta <= 1 && nCcpt+delta >= nSeq)) vecSize = 1 + ((n+1) * n/2);
        else if(nCcpt-delta <= 1 && nCcpt+delta <= nSeq) vecSize = 2 + ((nCcpt+delta) * (nSeq-1) - (nCcpt+delta-1)*(nCcpt+delta-2)/2);
        else if(nCcpt-delta > 1 && nCcpt+delta > nSeq) vecSize = 1 + (nSeq-nCcpt+delta+1)*(nSeq-nCcpt+delta+2)/2;
        else vecSize = 1 + ((2*delta + 1) * (nSeq+1 - nCcpt));

        assert(vecSize > 0 && vecSize <= matrix.max_size());     
        matrix = vector<double> (vecSize, numeric_limits<double>::max());    
    }

    double& operator ()(const int& j, const int& i, const int& nCcpt){
        assert(j >= 0 && i >= 0 && nCcpt >= 0 && j <= nSeq && i <= nSeq);
        assert(matrix[matrix.size()-1] == numeric_limits<double>::max());

        if(nCcpt-delta > nSeq) {
            return matrix[matrix.size()-1];
        } else if(nCcpt == 0 || (nCcpt-delta <= 1 && nCcpt+delta >= nSeq)){
            if(j >= i || i > nSeq) return matrix[matrix.size()-1];
            int memIndx = (nSeq+1)*j+i-((j+1)*(j+2))/2;
            assert(memIndx < matrix.size() && memIndx >= 0);
            return matrix[memIndx];
        } else if(nCcpt-delta <= 1 && nCcpt+delta < nSeq){
            if(j >= i || (i-j) > nCcpt + delta || (i-j) < nCcpt - delta){
                return matrix[matrix.size()-1];
            } else if(j > 0 && j > (nSeq-nCcpt-delta)){
                int memIndx = (nCcpt+delta)*j + (i-j-1) - (j-(nSeq-nCcpt-delta)-1)*(j-(nSeq-nCcpt-delta))/2;
                assert(memIndx < matrix.size()-1 && memIndx >= 0);
                return matrix[memIndx];
            } else {
                int memIndx = (nCcpt+delta)*j + (i-j-1);
                assert(memIndx < matrix.size()-1 && memIndx >= 0);
                return matrix[memIndx];
            }
        } else if(nCcpt-delta > 1 && nCcpt+delta > nSeq){
            if(j >= i || (i-j) > nCcpt + delta || (i-j) < nCcpt - delta){
                return matrix[matrix.size()-1];
            } else {
                int memIndx = (nSeq-nCcpt+delta+1)*j + i-(nCcpt-delta) - (j)*(j+1)/2;
                assert(memIndx < matrix.size()-1 && memIndx >= 0);
                return matrix[memIndx];
            }        
        } else {
            if(j >= i || (i-j) > nCcpt + delta || (i-j) < nCcpt - delta){
    			return matrix[matrix.size()-1];
    		} else if(j > 0 && j > (nSeq-nCcpt-delta)){
                int memIndx = (2*delta)*j + (i - (nCcpt-delta)) - ((j - (nSeq-nCcpt-delta)-1)*(j - (nSeq-nCcpt-delta)))/2;
                assert(memIndx < matrix.size()-1 && memIndx >= 0);
    			return matrix[memIndx];
    		} else {
                int memIndx = (2*delta)*j + (i - (nCcpt-delta));
                if(!(memIndx < matrix.size()-1 && memIndx >= 0)) cout << memIndx << " " << matrix.size() << " " << i << " " << j << " " << nCcpt << " " << nSeq << endl;
                assert(memIndx < matrix.size()-1 && memIndx >= 0);
    			return matrix[memIndx];
    		}
        }
    } 
};

#endif
