#ifndef DICTCLASS_H
#define DICTCLASS_H
#include <tuple>
#include <vector>
#include <algorithm>
#include <bits/stdc++.h>
#include <time.h> 
#include <cassert>
#include <map>
#include <list>
#include "util.h"
#include "cost.h"
using namespace std;

struct Concept {
    int _ind, _pos, _len;                   //Concept legend
    vector<Cost> costMat;                   //Vector of Cost matrices
    double conceptML;                       //vector of concept message lengths
    vector<tuple<int, int, int>> usage;     //Usage vector
};

class Dictionary{
private:
    list<Concept> conceptList;                        
    vector<vector<list<Concept>::iterator>> conceptLen;

    //Temporary variables for update/replace
    Concept tempCCPT;
    string logF;

    vector<string> dataSet;
    vector<double> mmlMatrix, condMatrix;

    double dictML = 0;
    int nSize = 0;
    int nSizeTrue = 0;
    double nullML = 0;
    int lowerBound;
    int upperBound;
    int updateLen;
    int delta;
    int epsln;
    int maxLen = 0;
    double regionLen = 0;

    //Flags
    bool disectLegend;      //data message length breakdown in getML, getMLDel
    bool pertLegend;        //perturbation legend in add,delete,update,replace
    bool disctBreakdown;    //dissection breakdown of each sequence
    bool prntCost;          //rod cutting cost and no. of dissections breakdown

    void populateCostMat(vector<Cost>&, const string&);
    void getHistoryMatrices(const string&, const string&, vector<vector<double>>&, vector<vector<double>>&, vector<vector<double>>&);
    void getBoundedHistoryMatrices(const string&, const string&, vector<vector<double>>&, vector<vector<double>>&, vector<vector<double>>&);
    void dissections(const string&, const int&, double&, int&);
    void updateConceptLength(const int&, int&, int&);
    
public:
    Dictionary(const vector<string>&, const int&, const int&, const int&, const int&, const int&, const int&, const double&, const vector<bool>&, vector<double>&, vector<double>&, const string&);
    void _pop();
    list<Concept>::iterator _randConcept();
    void _newConcept(int&, int&, int&);
    void _suppressConcept(const list<Concept>::iterator&);
    void _unsuppressConcept(const list<Concept>::iterator&);
    void _add(int&, int&, int&);
    void _remove(const list<Concept>::iterator&);
    void _update(list<Concept>::iterator&);
    bool _swap(list<Concept>::iterator&);  
    void _undoUpdateSwap(list<Concept>::iterator&);
    void _getML(double&, double&); 
    int _size();
    void _checkPoint(const string&);
    void _loadCheckPoint(const vector<string>&);
    void _printUsage(vector<string>&, const bool&, const bool&);   
};

#endif
