#include <bits/stdc++.h>
#include <math.h>
#include <stdio.h> 
#include "dictionary.h"
#include "cmdLineParser.h"
using namespace std;

string printLegend(double, double, int, double, int);

string logF = "logs/log-compass-vB-1.0.txt";

int main(int argc, char **argv){
    srand(0);

    //Command Line parser
    Parser input(argc, argv);

    //Populating dataset
    vector<string> dataSet; 
    double startKelv = input.kelvin;
    int maxSeq = 0;
    if(input.pflag) 
        file_append(logF, readSource(input.sourcePath, true, dataSet, input.lowerBound, input.seqLower, input.seqUpper, maxSeq));
    else 
        file_append(logF, readSource(input.sourceFile, false, dataSet, input.lowerBound, input.seqLower, input.seqUpper, maxSeq));
    
    if(input.regionLen == 0) input.regionLen = maxSeq;

    //Populating models (null and compression)
    vector<double> mmlMatrix(residueSize, 0), conditMatrix(residueSize*residueSize);
    init(input.nullModel, input.compModel, mmlMatrix, conditMatrix);

    //Initializing dictionary
    double dataml = 0, dictml = 0;
    vector<bool> passFlags = {input.disectLegend, input.pertLegend, input.seqDisect, input.seqCost};

    //Null model population
    Dictionary h(dataSet, input.lowerBound, input.upperBound, input.updateLen, input.delta, input.epsln, maxSeq, input.regionLen, passFlags, mmlMatrix, conditMatrix, logF);
    h._getML(dictml, dataml); 
    double acceptedML = dictml + dataml;
    double nullML = acceptedML;

    string tempParams = 
        "\n----------------------------- Parameters ----------------------------\n"
        "Temperature                          = " + num2str(input.kelvin, 6, 8) + "\n"
        "Cooling Temperature Probability      = " + num2str(input.kelvProb, 4, 8) + "\n"
        "Stopping Temperature                 = " + num2str(input.stoppingKelv, 3, 8) + "\n"
        "Number of Perturbations              = " + num2str(input.perturbations, 6, 8) + "\n"
        "Concept Minimum Length               = " + num2str(input.lowerBound, 3, 8) + "\n"
        "Delta                                = " + num2str(input.delta, 3, 8) + "\n"
        "Epsilon                              = " + num2str(input.epsln, 3, 8) + "\n"
        "---------------------------------------------------------------------\n";
    file_append(logF, tempParams);    
    cout << tempParams;

    string emptyDict = 
        "\n----------------------- Empty Dictionary Case -----------------------\n" 
        "Dictionary Cost                      = " + num2str(nits2bits(dictml), 6,11) + " bits\n" 
        "Data Cost                            = " + num2str(nits2bits(dataml), 10, 11) + " bits\n" 
        "Total Cost with empty dictionary     = " + num2str(nits2bits(nullML), 10, 11) + " bits\n" 
        "---------------------------------------------------------------------\n\n";
    cout << emptyDict;
    file_append(logF, emptyDict);
   
    //start with populated dictionary
    string populatedDict = "";
    if(input.cpData.size() > 0){
        h._loadCheckPoint(input.cpData);
        h._getML(dictml, dataml); 

        acceptedML = dataml + dictml; 
        assert(abs(acceptedML - input.checkpointedML) < 1e-6);
        if(!(abs(acceptedML - input.checkpointedML) < 1e-6)) {
            string checkPError = "Compass: Checkpoint message length error \n" 
                                "Previous Message Length = " + num2str(nits2bits(input.checkpointedML), 6,11) + "bits\n"
                                "Current Message Length  = " + num2str(nits2bits(acceptedML), 6,11) + "bits\n";
            cout << checkPError;
            file_append(logF, checkPError);  
            exit(0);                  
        }
        
        populatedDict = 
            "\n--------------------- Populated Dictionary Case ---------------------\n"
            "Dictionary Cost                      = " + num2str(nits2bits(dictml), 6,11) + " bits\n" 
            "Data Cost                            = " + num2str(nits2bits(dataml), 10, 11) + " bits\n" 
            "Total Cost with populated dictionary = " + num2str(nits2bits(acceptedML), 10, 11) + " bits\n" 
            "---------------------------------------------------------------------\n\n";
        cout << populatedDict;
        file_append(logF, populatedDict);
        vector<string> content;
        h._printUsage(content, true, false);
        system("mkdir usages");
        string fn = "usages/dissections_startingDictionary.txt";
        file_write(fn, content);
    }
    
    list<Concept>::iterator itIndx;
    string iterLegnd = "", criteria = "", type = "";
    int pertType = 0;
    int ind = -1, pos = -1, len = -1;
    int j;
    double newML = 0;
    while(input.kelvin >= input.stoppingKelv){
        for(j = input.iter; j < input.perturbations; j++){      	
            newML = 0;
            pertType = 0;
            ind = -1, pos = -1, len = -1;

            if(h._size() > 1){
                pertType = rand() % 4; 
               
                if(pertType == 0){ 
                    h._newConcept(ind, pos, len);
                    h._add(ind, pos, len);
                } else if(pertType == 1){
                    itIndx = h._randConcept();
                    h._suppressConcept(itIndx);
                } else if(pertType == 2){ 
                    h._getML(dictml, dataml); 
                    itIndx = h._randConcept();
                    bool retVal = h._swap(itIndx);
                    if(!retVal){ continue;} 
                } else {
                    itIndx = h._randConcept();
                    h._update(itIndx);
                }
            } else { 
                h._newConcept(ind, pos, len);
                h._add(ind, pos, len);
            }  

            h._getML(dictml, dataml); 
            newML =  dataml + dictml;
            
            populatedDict = 
                "--------------------- Populated Dictionary Case ---------------------\n"
                "Dictionary Cost                      = " + num2str(nits2bits(dictml), 6,11) + " bits\n" 
                "Data Cost                            = " + num2str(nits2bits(dataml), 10, 11) + " bits\n" 
                "Total Cost with populated dictionary = " + num2str(nits2bits(newML), 10, 11) + " bits\n" 
                "---------------------------------------------------------------------\n";
           
            double delta_obj = newML - acceptedML;
            
            switch(pertType){
                case 0: type = "Add"; break;
                case 1: type = "Del"; break;
                case 2: type = "Rep"; break;
                case 3: type = "Upd"; break;
            }
            
            iterLegnd = 
                num2str(input.kelvin, 8, 0) + "\t" + to_string(j) + "\t" + type + "\t" + num2str(nits2bits(newML), 9, 10) + "\t" + 
                num2str(nits2bits((nullML - newML)), 9, 10) + "\t" + num2str(nits2bits(-delta_obj), 9, 10) + "\t";
            
            if(delta_obj > -input.tolerance && delta_obj < input.tolerance) delta_obj = 0;     
            
            if(delta_obj < 0){
                acceptedML = newML;
                
                if (pertType == 1) 
                    h._remove(itIndx);
                
                criteria = "Accepted\n\n";

            } else {
                double prob = exp(-delta_obj/input.kelvin);

                double c = ((double) rand() / (RAND_MAX));
                assert(prob >=0 && prob <= 1);
                assert(c >=0 && c <= 1);
                
                if(prob > c){
                    acceptedML = newML;
                    
                    if (pertType == 1) 
                        h._remove(itIndx);

                    criteria = "Accepted with prob. " + num2str(prob, 5, 6) + "\n\n";
                
                } else {
                    if(pertType == 0) 
                        h._pop();  
                    else if(pertType == 1)
                        h._unsuppressConcept(itIndx);
                    else
                        h._undoUpdateSwap(itIndx);
                    
                    criteria = "Rejected\n\n";
                }
            }
            if(input.outFlag == 3) cout << populatedDict << iterLegnd << criteria;
            if(input.logFlag == 3) {
                file_append(logF, populatedDict);
                file_append(logF, iterLegnd);
                file_append(logF, criteria);
            }

            if(input.outFlag == 2 && j != 0 && j%1000 == 0) cout << populatedDict << iterLegnd << criteria;
            if(input.logFlag == 2 && j != 0 && j%1000 == 0){
                file_append(logF, populatedDict);
                file_append(logF, iterLegnd);
                file_append(logF, criteria);
            }
        }
        h._getML(dictml, dataml); 
        acceptedML = dictml + dataml;

        if(input.checkpoint){
            if(input.kelvin == startKelv) system("mkdir checkpoints");
            string f_name = "checkpoints/checkPointData_" + num2str(input.kelvin, 7, 0) + "_" + to_string(j) + ".txt";
            string content = to_string(input.kelvin) + " " + to_string(input.kelvProb) + " " + to_string(input.stoppingKelv) + " " + 
                             to_string(input.perturbations) + " " + to_string(input.lowerBound) + " " + to_string(input.upperBound) + " " + 
                             to_string(input.updateLen) + " " + to_string(j+1) + " " + to_string(input.delta) + " " + 
                             to_string(input.epsln) + " " + to_string(acceptedML) + "\n";             
            file_append(f_name, content);
            h._checkPoint(f_name);
        } 


        if(input.usageFile){
            if(input.kelvin == startKelv) system("mkdir usages");
            string f_name2 = "usages/dissections_" + num2str(input.kelvin, 7, 0)  + "_" + to_string(j) + ".txt";
            vector<string> content;
            string legend = printLegend(acceptedML, nullML, h._size(), input.kelvin, j);
            content.push_back(legend);
            file_append(logF, legend);
            h._printUsage(content, true, false);
            file_write(f_name2, content);
        }

        input.kelvin = (double) input.kelvin * input.kelvProb;
        input.iter = 0;


        if(input.outFlag >= 1) cout << populatedDict << iterLegnd << criteria;
        if(input.logFlag >= 1){
            file_append(logF, populatedDict);
            file_append(logF, iterLegnd);
            file_append(logF, criteria);
        } 
    }

    system("mkdir output");
    string f_name2 = "output/dictionary.txt";
    vector<string> content;

    string dataHeader = 
    "------------------------ Source Collection : ------------------------\n"
    "Total number of sequences considered : " + to_string(dataSet.size()) + 
    "---------------------------------------------------------------------\n";

    tempParams = 
    "\n----------------------------- Parameters ----------------------------\n"
    "Temperature                          = " + num2str(startKelv, 6, 8) + "\n"
    "Cooling Temperature Probability      = " + num2str(input.kelvProb, 4, 8) + "\n"
    "Stopping Temperature                 = " + num2str(input.stoppingKelv, 3, 8) + "\n"
    "Number of Perturbations              = " + num2str(input.perturbations, 6, 8) + "\n"
    "Epsilon                              = " + num2str(input.epsln, 3, 8) + "\n"
    "---------------------------------------------------------------------\n";
    content.push_back(dataHeader);
    content.push_back(tempParams);
    content.push_back(emptyDict);

    h._getML(dictml, dataml); 
    acceptedML = dictml + dataml;

    string legend = printLegend(acceptedML, nullML, h._size(), input.kelvin, j);
    content.push_back(legend);
    file_append(logF, legend);

    h._printUsage(content, true, false);
    file_write(f_name2, content);
}

string printLegend(double cost, double nullc, int ndict, double kelvin, int iter){
    string legend =  
        "------------------------- Dictionary Legend -------------------------\n"
        "Current cost                          : " + num2str(nits2bits(cost), 9, 10) + " bits\n" +
        "Initial cost                          : " + num2str(nits2bits(nullc), 9, 10) + " bits\n" +
        "Compression gain                      : " + num2str(nits2bits((nullc - cost)), 9, 10) + " bits\n\n" +
        "Number of dictionary elements         : " + num2str(ndict, 5, 10) + "\n\n" +
        "Current Temperature                   : " + num2str(kelvin, 10, 10) + "\n" +
        "Number of iterations                  : " + num2str(iter, 5, 10) + "\n" +
        "---------------------------------------------------------------------\n";
        
    cout << endl << endl << legend << "---------------------------------------------------------------------\n";    
    return legend;    
}
