#ifndef SCORE_H
#define SCORE_H
#include <vector>
#include <string>
#include <cassert>
#include <algorithm>
using namespace std;

class Score{
public:    
    vector<double> matrix;
    int nSeq;
    int nCcpt;
    int epsln;

    Score(const int& nn, const int& seql, const int& eps){
        nCcpt = nn;
        nSeq = seql;
        epsln = eps;
        assert(nCcpt> 0 && nSeq > 0 && epsln > 0);
        
        int base = (2*epsln+1)*(nSeq+1);
        int lft = triNum(epsln);
        int rgt = 0, vecSize = 0;
        assert(base > 0 && lft > 0);

        if(nSeq <= nCcpt){
            if(nSeq+epsln <= nCcpt) vecSize = 1 + base - lft;
            else if(epsln-nCcpt+nSeq >= 0) {
                rgt = triNum(epsln-nCcpt+nSeq);
                vecSize = 1 + base - lft - rgt;
            } else assert(false); 
        } else {
            if(nCcpt+epsln <= nSeq) vecSize = 1 + (2*epsln+1)*(nCcpt+1) - lft;
            else if(epsln-nCcpt+nSeq >= 0) {
                rgt = triNum(epsln-nCcpt+nSeq);
                vecSize = 1 + base - lft - rgt;
            } else assert(false);    
        }   
        if(epsln > min(nSeq,nCcpt)){
            if(nSeq < nCcpt) vecSize += triNum(epsln-nSeq-1);
            else vecSize += triNum(epsln-nCcpt-1);
        }
        assert(vecSize > 0 && vecSize <= matrix.max_size());   
        matrix = vector<double> (vecSize, numeric_limits<double>::max());
    }

    int triNum(int t){
        assert(t >= 0);
        return (t+1)*t/2;
    }

    double& operator ()(const int& i, const int& j){
        assert(i >= 0 && j >= 0 && j <= nCcpt && i <= nSeq);
        int exp1 = 0, exp2 = 0;
        if((epsln - (nCcpt-i) - 1) > 0) exp1 = triNum(epsln - (nCcpt-i) - 1);
        if(epsln > min(nSeq,nCcpt) && nCcpt <= nSeq && epsln-nCcpt >= 1) exp2 = triNum(epsln-nCcpt-1);
        
        if(i <= epsln && j <= epsln+i) {
            int memIndx = (2*epsln+1)*i+j - triNum(epsln) + triNum(epsln-i) - exp1 + exp2;
            if(!(memIndx < matrix.size()-1 && memIndx >= 0)) cout << i << " " << j << " " << memIndx << " " << nCcpt << " " << nSeq << " " << matrix.size() << endl;
            assert(memIndx < matrix.size()-1 && memIndx >= 0);
            return matrix[memIndx];
        } else if(i > epsln && j >= i-epsln && j <= i+epsln) {
            int memIndx = (2*epsln+1)*i+j - triNum(epsln) - (i-epsln) - exp1 + exp2;
            if(!(memIndx < matrix.size()-1 && memIndx >= 0)) cout << i << " " << j << " " << memIndx << " " << nCcpt << " " << nSeq << " " << matrix.size() << endl;
            assert(memIndx < matrix.size()-1 && memIndx >= 0);
            return matrix[memIndx];
        } 
        assert(matrix[matrix.size()-1] == numeric_limits<double>::max());
        return matrix[matrix.size() - 1]; 
    } 
};

#endif
